"""Vulture whitelist to avoid false positives."""


class Whitelist:
    """Helper class that allows mocking Python objects."""

    def __getattr__(self, _):
        """Mocking magic method __getattr__."""
        pass


whitelist_logging = Whitelist()
whitelist_logging.raiseExceptions

# Needed for vulture < 0.27
whitelist_mock = Whitelist()
whitelist_mock.return_value
whitelist_mock.side_effect

whitelist_tests = Whitelist()
whitelist_tests.unit.test_confctl.TestDns.setup_method



"""Sphinx configuration checker.

- Check that all the existing wmflib modules are listed in the API index documentation.
- Check that all the modules listed in the API index documentation exists in wmflib.
- Check that all the listed modules have its own file in the api/ documentation directory.
  Sphinx would raise a warning but not fail in this case.
"""
import argparse
import os
import pkgutil
import sys

import wmflib


DOC_API_BASE_PATH = 'doc/source/api'
DOC_API_INDEX_PATH = os.path.join(DOC_API_BASE_PATH, 'index.rst')
API_INDEX_PREFIX = '    wmflib.'
EXCLUDED_NAMES = ('log')


def main(base_path):
    """Perform the check."""
    wmflib_modules = {name for _, name, ispkg in pkgutil.iter_modules(wmflib.__path__)
                      if not ispkg and name not in EXCLUDED_NAMES}

    doc_path = os.path.join(base_path, DOC_API_INDEX_PATH)
    with open(doc_path) as f:
        api_index_lines = f.readlines()

    doc_api_lines = [line.strip() for line in api_index_lines if line.startswith(API_INDEX_PREFIX)]
    doc_api_modules = {line.split('.', 1)[1] for line in doc_api_lines}

    ret = 0
    if wmflib_modules - doc_api_modules:
        print('wmflib modules that are not listed in {doc}: {modules}'.format(
            doc=DOC_API_INDEX_PATH, modules=wmflib_modules - doc_api_modules))
        ret += 1
    if doc_api_modules - wmflib_modules:
        print('Documented modules in {doc} that are missing in wmflib: {modules}'.format(
            doc=DOC_API_INDEX_PATH, modules=doc_api_modules - wmflib_modules))
        ret += 1

    doc_api_files = ['wmflib.{name}.rst'.format(name=name) for name in doc_api_modules]
    missing_doc_api_files = [file for file in doc_api_files
                             if not os.path.isfile(os.path.join(DOC_API_BASE_PATH, file))]
    if missing_doc_api_files:
        print('Missing documentation files in {doc}: {files}'.format(
            doc=DOC_API_BASE_PATH, files=missing_doc_api_files))
        ret += 1

    if ret == 0:
        print('All wmflib modules are documented')

    return ret


if __name__ == '__main__':
    parser = argparse.ArgumentParser(  # pylint: disable=invalid-name
        description='Check that all wmflib modules are documented')
    parser.add_argument('base_path', help='Path to the root of the wmflib repository')
    args = parser.parse_args()  # pylint: disable=invalid-name

    sys.exit(main(args.base_path))



"""wmflib package."""

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__: str = get_distribution('wmflib').version  # Must be the same used as 'name' in setup.py
    """:py:class:`str`: the version of the current wmflib module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed



"""Exceptions module."""


class WmflibError(Exception):
    """Parent exception class for all wmflib exceptions."""


class WmflibCheckError(WmflibError):
    """Parent exception class for all wmflib exceptions regarding checks.

    Particularly useful when some write action is performed and then checked but in dry-run mode, as in this mode the
    write action will not actually change anything and the check will then fail, but should catchable separately from
    the other potential exceptions that could be raised.
    """



"""Package configuration."""

from setuptools import find_packages, setup


with open('README.rst', 'r') as readme:
    LONG_DESCRIPTION = readme.read()


INSTALL_REQUIRES = [
    'dnspython>=1.15.0',
]

# Extra dependencies
EXTRAS_REQUIRE = {
    # Test dependencies
    'tests': [
        'bandit>=1.1.0',
        'flake8>=3.2.1',
        'flake8-import-order>=0.18.1',
        'mypy>=0.470',
        # Temporary upper constraint until https://github.com/PyCQA/prospector/issues/389 is fixed
        'prospector[with_everything]>=0.12.4,<=1.2.0',
        'pytest-cov>=1.8.0',
        'pytest-xdist>=1.15.0',
        'requests-mock>=1.3.0',
        'sphinx_rtd_theme>=0.1.9',
        'sphinx-argparse>=0.1.15',
        'Sphinx>=1.4.9',
    ],
}

SETUP_REQUIRES = [
    'pytest-runner>=2.7.1',
    'setuptools_scm>=1.15.0',
]

setup(
    author='Luca Toscano',
    author_email='ltoscano@wikimedia.org',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Systems Administration',
    ],
    description='Generic library for common tasks in the WMF production infrastructure',
    entry_points={},
    extras_require=EXTRAS_REQUIRE,
    install_requires=INSTALL_REQUIRES,
    keywords=['wmf', 'automation'],
    license='GPLv3+',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    name='wmflib',
    packages=find_packages(exclude=['*.tests', '*.tests.*']),
    platforms=['GNU/Linux'],
    setup_requires=SETUP_REQUIRES,
    use_scm_version=True,
    url='https://github.com/wikimedia/operations-software-pywmflib',
    zip_safe=False,
)



"""DNS module."""
import logging

from typing import cast, List, Optional, Union

from dns import resolver, reversename, rrset
from dns.exception import DNSException
from dns.name import Name

from wmflib.exceptions import WmflibError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class DnsError(WmflibError):
    """Custom exception class for errors of the Dns class."""


class DnsNotFound(DnsError):
    """Custom exception class to indicate the record was not found.

    One or more resource records exist for this domain but there isn’t a record matching the resource record type.
    """


class Dns:
    """Class to interact with the DNS."""

    def __init__(self, *, nameserver_address: Optional[str] = None, port: Optional[int] = None) -> None:
        """Initialize the instance.

        Arguments:
            nameserver_address (str, optional): the nameserver address to use, if not set uses the OS configuration.
            port (int, optional): the port the ``nameserver_address`` nameserver is listening to, if different from
                the default 53. This applies only if a nameserver is explicitely specified.

        """
        if nameserver_address is not None:
            self._resolver = resolver.Resolver(configure=False)
            if port is not None:
                self._resolver.port = port
            self._resolver.nameservers = [nameserver_address]
        else:
            self._resolver = resolver.Resolver()

    def resolve_ipv4(self, name: str) -> List[str]:
        """Perform a DNS lookup for an A record for the given name.

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv4 addresses as strings returned by the DNS response.

        """
        return self._resolve_addresses(name, 'A')

    def resolve_ipv6(self, name: str) -> List[str]:
        """Perform a DNS lookup for an AAAA record for the given name.

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv6 addresses as strings returned by the DNS response.

        """
        return self._resolve_addresses(name, 'AAAA')

    def resolve_ips(self, name: str) -> List[str]:
        """Perform a DNS lookup for A and AAAA records for the given name.

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv4 and IPv6 addresses as strings returned by the DNS response.

        Raises:
            wmflib.dns.DnsNotFound: when no address is found.

        """
        addresses = []
        for func in ('resolve_ipv4', 'resolve_ipv6'):
            try:
                addresses += getattr(self, func)(name)
            except DnsNotFound:
                pass  # Allow single stack answers

        if not addresses:
            raise DnsNotFound('Record A or AAAA not found for {name}'.format(name=name))

        return addresses

    def resolve_ptr(self, address: str) -> List[str]:
        """Perform a DNS lookup for PTR record for the given address.

        Arguments:
            address (str): the IPv4 or IPv6 address to resolve.

        Returns:
            list: the list of absolute target PTR records as strings, without the trailing dot.

        """
        response = self.resolve(reversename.from_address(address), 'PTR')
        return self._parse_targets(cast(rrset.RRset, response.rrset))

    def resolve_cname(self, name: str) -> str:
        """Perform a DNS lookup for CNAME record for the given name.

        Arguments:
            name (str): the name to resolve.

        Returns:
            str: the absolute target name for this CNAME, without the trailing dot.

        """
        targets = self._parse_targets(cast(rrset.RRset, self.resolve(name, 'CNAME').rrset))
        if len(targets) != 1:
            raise DnsError('Found multiple CNAMEs target for {name}: {targets}'.format(name=name, targets=targets))

        return targets[0]

    def resolve(self, qname: Union[str, Name], record_type: str) -> resolver.Answer:
        """Perform a DNS lookup for the given qname and record type.

        Arguments:
            qname (str): the name or address to resolve.
            record_type (str): the DNS record type to lookup for, like 'A', 'AAAA', 'PTR', etc.

        Returns:
            dns.resolver.Answer: the DNS response.

        Raises:
            wmflib.dns.DnsNotFound: if there are no records for the given record type but the qname has records for
                different record type(s).
            wmflib.dns.DnsError: on generic error.

        """
        try:
            response = self._resolver.query(qname, record_type)
            logger.debug('Resolved %s record for %s: %s', record_type, qname, response.rrset)
        except (resolver.NoAnswer, resolver.NXDOMAIN) as e:
            raise DnsNotFound('Record {record_type} not found for {qname}'.format(
                record_type=record_type, qname=qname)) from e
        except DNSException as e:
            raise DnsError('Unable to resolve {record_type} record for {qname}'.format(
                record_type=record_type, qname=qname)) from e

        return response

    def _resolve_addresses(self, name: str, record_type: str) -> List[str]:
        """Extract and return all the matching addresses for the given name and record type.

        Arguments:
            name (str): the name to resolve.
            record_type (str): the DNS record type to lookup for, like 'A' and 'AAAA'.

        Returns:
            list: the list of IPv4 or IPv6 addresses as strings returned by the DNS response.

        """
        return [rdata.address for rdata in cast(rrset.RRset, self.resolve(name, record_type).rrset)]

    @staticmethod
    def _parse_targets(response_set: rrset.RRset) -> List[str]:
        """Extract and return all the matching names from the given rrset without the trailing dot.

        Arguments:
            response_set (dns.rrset.RRset): the RRset to parse.

        Returns:
            list: the list of absolute target record names as strings without the trailing dot.

        Raises:
            wmflib.dns.DnsError: if a relative record is found.

        """
        targets = []
        for rdata in response_set:
            target = rdata.target.to_text()
            if target[-1] != '.':
                raise DnsError('Unsupported relative target {target} found'.format(target=target))

            targets.append(target[:-1])

        return targets



"""DNS module tests."""
from collections import namedtuple
from unittest import mock

import dns
import pytest

from wmflib.dns import Dns, DnsError, DnsNotFound


# TODO: convert the mocked objects using dnspython objects. It requires quite some code given the structure of
#       dnspython API.
MockedDnsAddress = namedtuple('MockedDnsAddress', ['address'])
MockedDnsTarget = namedtuple('MockedDnsTarget', ['target'])
MockedDnsAnswer = namedtuple('MockedDnsAnswer', ['ttl', 'rrset'])


class MockedTarget:
    """Represent a DNS PTR response target."""

    def __init__(self, target):
        """Constructor."""
        self.target = target

    def to_text(self):
        """Required to mock the dnspython object."""
        return self.target


MOCKED_RESPONSES = {
    # (qname, record_type): MockedDnsAnswer(),
    ('host1.example.com', 'A'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsAddress(address='10.0.0.1')]),
    ('host1.example.com', 'AAAA'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsAddress(address='2001::1')]),
    ('10.0.0.1', 'PTR'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsTarget(target=MockedTarget('host1.example.com.'))]),
    ('2001::1', 'PTR'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsTarget(target=MockedTarget('host1.example.com.'))]),
    ('host2.example.com', 'A'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsAddress(address='10.0.0.1')]),
    ('host2.example.com', 'AAAA'): dns.resolver.NoAnswer('Not found'),
    ('host3.example.com', 'A'):
        MockedDnsAnswer(ttl=600, rrset=[MockedDnsAddress(address='10.0.0.1'), MockedDnsAddress(address='10.0.0.2')]),
    ('host3.example.com', 'AAAA'):
        MockedDnsAnswer(ttl=600, rrset=[MockedDnsAddress(address='2001::1'), MockedDnsAddress(address='2001::2')]),
    ('2001::2', 'PTR'):
        MockedDnsAnswer(ttl=600, rrset=[
            MockedDnsTarget(target=MockedTarget('host3.example.com.')),
            MockedDnsTarget(target=MockedTarget('service.example.com.'))]),
    ('service.example.com', 'CNAME'): MockedDnsAnswer(
        ttl=600, rrset=[MockedDnsTarget(target=MockedTarget('host1.example.com.'))]),
    ('multiservice.example.com', 'CNAME'): MockedDnsAnswer(
        ttl=600, rrset=[MockedDnsTarget(target=MockedTarget('host1.example.com.')),
                        MockedDnsTarget(target=MockedTarget('host2.example.com.'))]),
    ('relative.example.com', 'CNAME'): MockedDnsAnswer(ttl=600, rrset=[MockedDnsTarget(target=MockedTarget('host1'))]),
}


def mocked_dns_query(qname, record_type):
    """Mock a dnspython query response."""
    if record_type == 'PTR':
        qname = dns.reversename.to_address(qname)
        if isinstance(qname, bytes):
            qname = qname.decode()

        if qname.startswith('192.168.') or qname.startswith('fe80::'):
            raise dns.exception.DNSException('Not defined')

    elif record_type in ('A', 'AAAA') and qname.startswith('raise.'):
        raise dns.exception.DNSException('Not defined')
    elif record_type in ('A', 'AAAA') and qname.startswith('missing.'):
        raise dns.resolver.NoAnswer('Not found')

    response = MOCKED_RESPONSES[(qname, record_type)]
    if isinstance(response, Exception):
        raise response

    return response


class TestDns:
    """Dns class tests."""

    @mock.patch('wmflib.dns.resolver.Resolver')
    def setup_method(self, _, mocked_resolver):
        """Initialize the test environment for Dns."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_resolver = mocked_resolver
        self.mocked_resolver.return_value.query = mocked_dns_query
        self.dns = Dns()

    def test_init(self):
        """The dns.resolver.Resolver should have been called without parameters."""
        self.mocked_resolver.assert_called_once_with()

    @mock.patch('wmflib.dns.resolver.Resolver')
    def test_init_with_nameserver(self, mocked_resolver):
        """When passing a nameserver address, this should be set in the dns Resolver too."""
        Dns(nameserver_address='127.0.0.1')
        mocked_resolver.assert_called_once_with(configure=False)
        assert mocked_resolver.return_value.nameservers == ['127.0.0.1']
        self.mocked_resolver.assert_called_once_with()

    @mock.patch('wmflib.dns.resolver.Resolver')
    def test_init_with_nameserver_and_port(self, mocked_resolver):
        """A non-standard port should be set in the dns Resolver if a nameserver is set."""
        Dns(nameserver_address='127.0.0.1', port=5353)
        mocked_resolver.assert_called_once_with(configure=False)
        assert mocked_resolver.return_value.nameservers == ['127.0.0.1']
        assert mocked_resolver.return_value.port == 5353
        self.mocked_resolver.assert_called_once_with()

    def test_resolve_ipv4(self):
        """Should return the list of IPv4 matching the name."""
        assert self.dns.resolve_ipv4('host1.example.com') == ['10.0.0.1']
        assert self.dns.resolve_ipv4('host3.example.com') == ['10.0.0.1', '10.0.0.2']

    def test_resolve_ipv6(self):
        """Should return the list of IPv6 matching the name."""
        assert self.dns.resolve_ipv6('host1.example.com') == ['2001::1']
        assert self.dns.resolve_ipv6('host3.example.com') == ['2001::1', '2001::2']

    def test_resolve_ips_ok(self):
        """Should return the list of IPv4 and IPv6 matching the name."""
        assert self.dns.resolve_ips('host1.example.com') == ['10.0.0.1', '2001::1']
        assert self.dns.resolve_ips('host3.example.com') == ['10.0.0.1', '10.0.0.2', '2001::1', '2001::2']

    def test_resolve_ips_one_only(self):
        """Should not fail there is no match for IPv4 or IPv6."""
        assert self.dns.resolve_ips('host2.example.com') == ['10.0.0.1']

    def test_resolve_ips_not_found(self):
        """Should raise DnsNotFound exception if the name raises NoAnswer for both A and AAAA records."""
        with pytest.raises(DnsNotFound, match='Record A or AAAA not found for missing.example.com'):
            self.dns.resolve_ips('missing.example.com')

    @pytest.mark.parametrize('address, response', (
        ('10.0.0.1', ['host1.example.com']),
        ('2001::1', ['host1.example.com']),
        ('2001::2', ['host3.example.com', 'service.example.com']),
    ))
    def test_resolve_ptr(self, address, response):
        """Should return the list of pointers matching the address."""
        assert self.dns.resolve_ptr(address) == response

    def test_resolve_cname(self):
        """Should return the target of a CNAME."""
        assert self.dns.resolve_cname('service.example.com') == 'host1.example.com'

    def test_resolve_cname_multiple_targets(self):
        """Should raise DnsError if more than one target is returned."""
        with pytest.raises(DnsError, match='Found multiple CNAMEs target for multiservice.example.com'):
            self.dns.resolve_cname('multiservice.example.com')

    def test_resolve_cname_relative_target(self):
        """Should raise DnsError if a relative target is found."""
        with pytest.raises(DnsError, match='Unsupported relative target host1 found'):
            self.dns.resolve_cname('relative.example.com')

    def test_resolve_not_found(self):
        """Should raise DnsNotFound exception if the record type is not defined for the qname."""
        with pytest.raises(DnsNotFound, match='Record AAAA not found for host2.example.com'):
            self.dns.resolve('host2.example.com', 'AAAA')

    @pytest.mark.parametrize('qname, record_type', (
        ('raise.example.com', 'A'),
        ('raise.example.com', 'AAAA'),
        (dns.reversename.from_address('192.168.1.1'), 'PTR'),
        (dns.reversename.from_address('fe80::1'), 'PTR'),
    ))
    def test_resolve_raise(self, qname, record_type):
        """Should raise DnsError if the qname is not defined."""
        with pytest.raises(DnsError, match='Unable to resolve {record_type} record for {qname}'.format(
                record_type=record_type, qname=qname)):
            self.dns.resolve(qname, record_type)
